﻿using System;
using System.Collections.Generic;
using System.Drawing;

namespace TestProject
{
    /// <summary>
    /// 그래픽스 확장
    /// </summary>
    public static class GraphicsExtension
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 난수기
        /// </summary>
        private static Random _random = new Random();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 포인트 그리기 - DrawPoint(graphics, point, brush, pen, radius)

        /// <summary>
        /// 포인트 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="point">포인트</param>
        /// <param name="brush">브러시</param>
        /// <param name="pen">펜</param>
        /// <param name="radius">반경</param>
        public static void DrawPoint(this Graphics graphics, PointF point, Brush brush, Pen pen, float radius)
        {
            RectangleF rectangle = new RectangleF
            (
                point.X - radius,
                point.Y - radius,
                2 * radius,
                2 * radius
            );

            graphics.FillEllipse(brush, rectangle);

            graphics.DrawEllipse(pen, rectangle);
        }

        #endregion
        #region 포인트 리스트 그리기 - DrawPointList(graphics, pointList, brush, pen, radius)

        /// <summary>
        /// 포인트 리스트 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="pointList">포인트 리스트</param>
        /// <param name="brush">브러시</param>
        /// <param name="pen">펜</param>
        /// <param name="radius">반경</param>
        public static void DrawPointList(this Graphics graphics, List<PointF> pointList, Brush brush, Pen pen, float radius)
        {
            if(pointList == null)
            {
                return;
            }

            foreach(PointF point in pointList)
            {
                graphics.DrawPoint(point, brush, pen, radius);
            }
        }

        #endregion
        #region 사각형 그리기 - DrawRectangle(graphics, pen, rectangle)

        /// <summary>
        /// 사각형 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="pen">펜</param>
        /// <param name="rectangle">사각형</param>
        public static void DrawRectangle(this Graphics graphics, Pen pen, RectangleF rectangle)
        {
            graphics.DrawRectangle(pen, rectangle.X, rectangle.Y, rectangle.Width, rectangle.Height);
        }

        #endregion
        #region 사각형 그리기 - DrawRectangle(graphics, point, brush, pen, radius)

        /// <summary>
        /// 사각형 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="point">포인트</param>
        /// <param name="brush">브로시</param>
        /// <param name="pen">펜</param>
        /// <param name="radius">반경</param>
        public static void DrawRectangle(this Graphics graphics, PointF point, Brush brush, Pen pen, float radius)
        {
            RectangleF rectangle = new RectangleF
            (
                point.X - radius,
                point.Y - radius,
                2 * radius,
                2 * radius
            );

            graphics.FillRectangle(brush, rectangle);

            graphics.DrawRectangle(pen, rectangle);
        }

        #endregion
        #region 십자가 그리기 - DrawCross(graphics, pen, point, radius)

        /// <summary>
        /// 십자가 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="pen">펜</param>
        /// <param name="point">포인트</param>
        /// <param name="radius">반경</param>
        public static void DrawCross(this Graphics graphics, Pen pen, PointF point, float radius)
        {
            graphics.DrawLine(pen, point.X - radius, point.Y, point.X + radius, point.Y);

            graphics.DrawLine(pen, point.X, point.Y - radius, point.X, point.Y + radius);
        }

        #endregion
        #region 십자가 그리기 - DrawCross(graphics, outerColor, innerColor, point, radius)

        /// <summary>
        /// 십자가 그리기
        /// </summary>
        /// <param name="graphics">그래픽스</param>
        /// <param name="outerColor">외부 색상</param>
        /// <param name="innerColor">내부 색상</param>
        /// <param name="point">포인트</param>
        /// <param name="radius">반경</param>
        public static void DrawCross(this Graphics graphics, Color outerColor, Color innerColor, PointF point, float radius)
        {
            using(Pen pen = new Pen(outerColor, 3))
            {
                graphics.DrawLine(pen, point.X - radius - 1, point.Y, point.X + radius + 1, point.Y);

                graphics.DrawLine(pen, point.X, point.Y - radius - 1, point.X, point.Y + radius + 1);
            }

            using(Pen pen = new Pen(innerColor, 1))
            {
                graphics.DrawLine(pen, point.X - radius, point.Y, point.X + radius, point.Y);

                graphics.DrawLine(pen, point.X, point.Y - radius, point.X, point.Y + radius);
            }
        }

        #endregion

        #region 난수 구하기 - Random<T>(list)

        /// <summary>
        /// 난수 구하기
        /// </summary>
        /// <typeparam name="TItem">항목 타입</typeparam>
        /// <param name="list">리스트</param>
        /// <returns>난수 값</returns>
        public static TItem Random<TItem>(List<TItem> list)
        {
            return list[_random.Next(list.Count)];
        }

        #endregion
        #region 임의로 섞기 - Randomize<T>(array)

        /// <summary>
        /// 임의로 섞기
        /// </summary>
        /// <typeparam name="TItem">항목 타입</typeparam>
        /// <param name="array">배열</param>
        public static void Randomize<TItem>(this TItem[] array)
        {
            for(int i = 0; i < array.Length - 1; i++)
            {
                int j = _random.Next(i, array.Length);

                TItem temporaryValue = array[i];

                array[i] = array[j];
                array[j] = temporaryValue;
            }
        }

        #endregion
        #region 임의로 섞기 - Randomize<T>(list)

        /// <summary>
        /// 임의로 섞기
        /// </summary>
        /// <typeparam name="TItem">항목 타입</typeparam>
        /// <param name="list">리스트</param>
        public static void Randomize<TItem>(this List<TItem> list)
        {
            TItem[] array = list.ToArray();

            array.Randomize();

            list.Clear();

            list.AddRange(array);
        }

        #endregion
    }
}